/*
   m3_Hex2Uc.c
   class to program program a m3 microcontroller's flash memory
   over the ethernet.

   finucane@myri.com (David Finucane)
*/

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <stdarg.h>
#include <ctype.h>

#include <sys/types.h>
#include <sys/socket.h>
#include <sys/time.h>
#include <netinet/in.h>
#include <netdb.h>
#include <arpa/inet.h>
#include <unistd.h>

#include "m3_Hex2Uc.h"
#include "insist.h"

#define MAX_LINE 100
#define MAX_TERMINAL_MESSAGE_LENGTH 10000
#define TYPE_MARK "Type: "



/* message format:

   M3_UC_FLASH_STRING
   4 byte flag = 0
   4 byte port
   4 byte size
   image

 */

m3_Hex2Uc::m3_Hex2Uc (int verbose)
{
  this->verbose = verbose;
}


int m3_Hex2Uc::check_buf (char*buf)
{
  insist (this && buf);
  
  int i;

  if (verbose) printFormat("checking address ranges");

  for(i=0; i<=0x7FFF; i++) {
    if (buf[i] != -1) {
      printFormat("Error, S19 file has data at address %04x.",i);
      printFormat("A 68HC908GP32 doesn't have flash at this location");
      return 0;
    }
  }
  for(i=0xFE00; i<=0xFF7D; i++) {
    if (buf[i] != -1) {
      printFormat("Error, S19 file has data at address %04x.",i);
      printFormat("A 68HC908GP32 doesn't have flash at this location");
      return 0;
    }
  }
  for(i=0xFF7F; i<=0xFFDB; i++) {
    if (buf[i] != -1) {
      printFormat("Error, S19 file has data at address %04x.",i);
      printFormat("A 68HC908GP32 doesn't have flash at this location");
      return 0;
    }
  }
  if (verbose) printFormat("complete.");
  return 1;
  exception: return 0;
}


int m3_Hex2Uc::read_s19 (FILE*f, char*buf) 
{
  char line[256];
  int len;
  int addr;
  int i;
  int mysum, checksum;
  char *c;
  int line_number = 0;

  insist (this && f && buf);
  
  memset (buf, -1,M3_UC_FLASH_MAX_IMAGE);
  
  if (verbose) printFormat("reading s-record file.");
  for (;;)
  {
    line_number++;
    
    if (!fgets(line,256,f))
    {
      printFormat ("premature end of file at line %d", line_number);
      return 0;
    }
    
    if (!strncmp(line,"S0",2))
      continue;
    
    if (!strncmp(line,"S1",2))
    {
      c = line+2;
      len =-1;
      if (sscanf(c,"%2x",&len) != 1)
      {
	printFormat ("premature end of file at length field line %d", line_number);
	return 0;
      }
	
      if (len == -1) 
      {
        printFormat("Error with length field, line %d", line_number);
        return 0;
      }
      mysum = len;
      c +=2;
      if (sscanf(c,"%4x",&addr) != 1)
      {
	printFormat ("premature end of file at address, line %d", line_number);
	return 0;
      }
      mysum += addr&0xFF;
      mysum += addr>>8;
      c+=4;
      for(i=0; i<len-3; i++) 
      {
	int data;
        if (sscanf(c,"%2x",&data) != 1)
	{
	  printFormat ("premature end of file at data, line %d", line_number);
	  return 0;
	}
	
        c +=2;
        mysum += data;
	buf [addr+i] = data & 0xff;
      }

      if (sscanf(c,"%2x",&checksum) != 1)
      {
	printFormat ("premature end of file at checksum, line %d", line_number);
	return 0;
      }
      if (checksum != ((mysum&0xFF)^0xFF)) 
      {
        printFormat("Checksum error on line:%d",line_number);
        return 0;
      }
    } 
    else if (!strncmp(line,"S9", 2)) 
      break;
    if (feof(f)) 
    {
      printFormat("Missing S9 termination record, line %d", line_number);
      return 0;
    }
  }
  if (verbose) printFormat ("complete.");
  return 1;
  exception: return 0;
}

int m3_Hex2Uc::port_connect (char *hostname, int port)
{
  struct hostent *he;
  struct sockaddr_in sin;
  int fd;

  insist (this && hostname && port > 0);
  
  memset(&sin, 0, sizeof(sin));
  sin.sin_family = AF_INET;
  sin.sin_port = htons(port);

  if (he = gethostbyname(hostname))
    memcpy((char *) &sin.sin_addr, he->h_addr, he->h_length);
  else if ((sin.sin_addr.s_addr = inet_addr(hostname)) == -1) 
  {
    printFormat ("can't get IP address of %s", hostname);
    return 0;
  }

  if ((fd = socket(PF_INET, SOCK_STREAM, 0)) < 0) 
  {
    perror("socket");
    return 0;
  }

  if (connect(fd, (struct sockaddr *) &sin, sizeof(sin)) < 0) 
  {
    perror("connect");
    return 0;
  }
  return fd;
  exception: return 0;
}

char terminal_buffer[MAX_TERMINAL_MESSAGE_LENGTH];

int m3_Hex2Uc::epilogue (int fd)
{
  int c;
  char*s;
  
  for (;;) 
  {
    memset (terminal_buffer, 0, MAX_TERMINAL_MESSAGE_LENGTH);
    c = read (fd, terminal_buffer, MAX_TERMINAL_MESSAGE_LENGTH);
    if (c <= 0)
      return 1;

    for (s = terminal_buffer; s < terminal_buffer + c; s += strlen(s) + 1)
      printFormat ("%s", s);
    fflush (stdout);
  }
  return 1;
}

int m3_Hex2Uc::program (char*hostname, char*filename, int port)
{  
  int fd, c;
  FILE *fp;
  int size = M3_UC_FLASH_MAX_IMAGE;
  int flag = 0;
  int bytes;
  int tb, i;

  insist (this);
  insist (hostname && filename);
  insist (port >= 0 && port < M3_MAX_SERIAL_PORT);  

  memset (image, 0, size);

  if (port < 0 || port > M3_MAX_SERIAL_PORT)
  {
    printFormat ("bad port number %d", port);
    return 0;
  }
  
  if (!(fp = fopen (filename, "r"))) 
  {
    printFormat ("couldn't open %s", filename);
    return 0;
  }

  printFormat ("reading image file.");
  if (!read_s19 (fp, image))
  {
    printFormat ("error in %s", filename);
    fclose (fp);
    return 0;
  }

  fclose (fp);
  if (!check_buf (image))
  {
    printFormat ("error in %s", filename);
    return 0;
  }
  
  if (!(fd = port_connect (hostname, M3_UC_FLASH_PORT))) 
  {
    printFormat ("couldn't connect to %s:%d.", hostname, M3_UC_FLASH_PORT);
    return 0;
  }
  
  printFormat ("writing FLASH STRING");
  if (write(fd, (void*) M3_UC_FLASH_STRING, strlen (M3_UC_FLASH_STRING)) != strlen (M3_UC_FLASH_STRING)) 
  {
    printFormat ("couldn't write FLASH_STRING.");
    return epilogue (fd);;
  }

  printFormat ("writing flag %d", flag);
  tb = htonl (flag);
  if (write (fd, &tb, sizeof(int)) != sizeof (int)) 
  {
    printFormat ("couldn't write flag.");
    return epilogue (fd);;
  }
  
  printFormat ("writing port %x", port);
  tb = htonl (port);
  if (write (fd, &tb, sizeof(int)) != sizeof(int)) 
  {
    printFormat ("couldn't write port.");
    return epilogue (fd);;
  }

  printFormat ("writing size");
  tb = htonl (size);
  if (write (fd, &tb, sizeof(int)) != sizeof(int)) 
  {
    printFormat ("couldn't write size.");
    return epilogue (fd);;
  }

  printFormat ("writing image");
  fflush (stdout);
  if (write (fd, image, size) != size) 
  {
    printFormat ("couldn't write image.");
    return epilogue (fd);;
  }

  printFormat ("done.");

  return epilogue (fd);
  exception: return 0;
}

int m3_Hex2Uc::getFirmwareVersion (char*filename, char*firmwareVersion)
{
  FILE*fp;
  
  insist (this && filename && firmwareVersion);

  memset (image, 0, M3_UC_FLASH_MAX_IMAGE);

  if (!(fp = fopen (filename, "r"))) 
  {
    printFormat ("couldn't open %s", filename);
    return 0;
  }

  if (!read_s19 (fp, image) || !check_buf (image))
  {
    printFormat ("error in %s", filename);
    fclose (fp);
    return 0;
  }
  fclose (fp);


  for (char*s = image; s < image + M3_UC_FLASH_MAX_IMAGE; s += 1 + strlen (s))
  {
    char*t;
    
    if ((t = strstr (s, TYPE_MARK)))
    {
      char*f = t;
      for (int i = 0; i < 3; i++)
      {
	while (*f && f < image + M3_UC_FLASH_MAX_IMAGE)
	  f++;
	*f = '\n';
      }
	     
      strcpy (firmwareVersion, t);
      return 1;
    }
  }
  return 1;
  exception: return 0;
}
