/*
  test_port.c
  program to talk to switch through test port
  finucane@myri.com (David Finucane)
*/

#include <stdio.h>
#include <stdarg.h>
#include <stdlib.h>
#include <time.h>
#include <string.h>
#include <signal.h>

#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <unistd.h>
#include <netdb.h>

#include "m3_defines.h"
#include "m3_TestClient.h"
#include "m3_RollcallClient.h"


m3_int64 m2_good [16*16];
m3_int64 m2_bad [16*16];
int invalid [16 * 16];

char*string64 (char*s, unsigned low, unsigned high)
{
#ifdef m3_intel_linux
  unsigned long long a = ((unsigned long long)low) | ((unsigned long long) high << 32);
  sprintf (s, "%llu", a);
#else
  if (high)
    sprintf (s, "0x%X%08X", high, low);
  else
    sprintf (s, "%u", low);
#endif
  return s;
}

inline unsigned long long htonll (unsigned long long x)
{
  unsigned a, b;
  unsigned long long c;

  if (htonl (1) == 1) return x;
  
  a = x;
  b = x >> 32;
  c = htonl (b) | htonl (a);
  return c;
}


static void printLine (char*s)
{
  printf ("%s\n", s);
}

static m3_TestClient*client = 0;
static void signal_handler (int)
{
  if (client)
    client->signaled = 1;
}

static void usage (char*s)
{
  if (s)
    printf ("unknown option %s\n", s);
  
  printf ("usage:\n");
  printf ("test_port [host]\n");
  printf (" -get-version-string\n");
  printf (" -reboot\n");
  printf (" -get-uptime\n");
  printf (" -get-extended-values\n");
  printf (" -get-ethernet-address\n");
  printf (" -get-active-ethernet-port\n");
  printf (" -set-active-ethernet-port <port>\n");
  printf (" -serial-test\n");
  printf (" -stop-update\n");
  printf (" -start-update\n");
  printf (" -debug\n");
  printf (" -test-link <mtu>\n");
  printf (" -notify <type_bits_1_32> [<type_bits_33_64>]\n");
  printf (" -set-web-password [<username> <password>]\n");
  printf (" -lock\n");
  printf (" -disable <taskname>\n");
  printf (" -get-packet-counts\n");
  printf (" -get-m2-packet-counts\n");
  exit (1);
}

static void fail (char*s)
{
  printf ("%s failed\n", s);
  exit (1);
}

int main (int argc, char**argv)
{
  m3_Component::initialize (printLine);
  m3_RollcallClient rollcall;
  int*switches = 0;
  int numSwitches = 0;
  
  if (argc < 2)
    usage (0);
 
  client = new m3_TestClient ();

  if (argc == 2 || *argv [1] == '-')
  {
    rollcall.open ();
    numSwitches = rollcall.findAddresses (&switches);
    rollcall.close ();
    printf ("%d m3 switches\n", numSwitches);
  }
  
  
  for (int times = 0;;times++)
  {
    if (numSwitches && strcmp (argv[1], "-notify"))
    {
      if (times == numSwitches)
	return 0;
      if (!client->open (switches [times]))
	fail ("open");

      printf ("\n%d.%d.%d.%d\n",
	      (unsigned) (((char*)(&switches [times])) [0] & 0xff),
	      (unsigned) (((char*)(&switches [times])) [1] & 0xff),
	      (unsigned) (((char*)(&switches [times])) [2] & 0xff),
	      (unsigned) (((char*)(&switches [times])) [3] & 0xff));
    }
    else if (!numSwitches)
    {
      if (times)
	return 0;
      
      if (!client->open (argv [1]))
	fail ("open");
    }
    
    for (int i = numSwitches ? 1 : 2; i < argc; i++)
    {
      if (!strcmp (argv [i], "-get-version-string"))
      {
	char version [M3_TEST_VERSION_STRING_LENGTH + 1];

	if (!client->getVersionString (version))
	  fail (argv [i]);
      
	printf ("%s\n", version);
      }
      else if (!strcmp (argv [i], "-reboot"))
      {
	if (!client->reboot ())
	  fail (argv [i]);
      }
      else if (!strcmp (argv [i], "-get-uptime"))
      {
	int e;
	printf ("%d second%s\n", (e = client->getUptime ()), e == 1 ? "" :"s" );
      }
      else if (!strcmp (argv [i], "-get-extended-values"))
      {
	int numValues = client->getExtendedCount ();
	for (int j = 0; j < numValues; j++)
	  printf ("%d. %d\n", j, client->getExtendedValue (j));
      }
      else if (!strcmp (argv [i], "-get-ethernet-address"))
      {
	char address [6];

	if (!client->getEthernetAddress (address))
	  fail (argv [i - 1]);
      
	printf ("%2.2x:%2.2x:%2.2x:%2.2x:%2.2x:%2.2x\n",
		0xff & address [0],
		0xff & address [1],
		0xff & address [2],
		0xff & address [3],
		0xff & address [4],
		0xff & address [5]);
      }
      else if (!strcmp (argv [i], "-get-active-ethernet-port"))
      {
	printf ("port %d is active.\n", client->getActiveEthernetPort ());
      }
      else if (!strcmp (argv [i], "-set-active-ethernet-port"))
      {
	if (++i == argc)
	{
	  printf ("expected ethernet port\n");
	  usage (0);
	}
	if (!client->setActiveEthernetPort (atoi (argv [i])))
	  fail (argv [i - 1]);
      
	printf ("port %d is active.\n", client->getActiveEthernetPort ());
      }
      else if (!strcmp (argv [i], "-debug"))
      {
	if (!client->debug ())
	  fail (argv [i]);
      }
      else if (!strcmp (argv [i], "-test-link"))
      {
	signal (SIGTERM, signal_handler);
	signal (SIGINT, signal_handler);
  
	if (++i == argc)
	{
	  printf ("expected mtu\n");
	  usage (0);
	}
	int mtu;
	mtu = atoi (argv [i]);

	if (mtu > M3_TEST_MTU || mtu < sizeof (m3_test_link_header_t))
	{
	  printf ("mtu must be from %d to %d\n", sizeof (m3_test_link_header_t), M3_TEST_MTU);
	  usage (0);
	}

	time_t a, b;
	int received;
	a = time (0);
	received = client->testLink (mtu);
	b = time (0);

	if (a == b || received == 0)
	  fail (argv [i - 1]);
      
	printf ("%d received. %2.2f mbps\n", received, 
		(((double) mtu * (double) 8 * (double) received / difftime (b, a)) / (double)(1000 * 1000)));
      
      }
      else if (!strcmp (argv [i], "-start-update"))
      {
	if (!client->startUpdate ())
	  fail (argv [i]);
      }
      else if (!strcmp (argv [i], "-stop-update"))
      {
	if (!client->stopUpdate ())
	  fail (argv [i]);
      }
      else if (!strcmp (argv [i], "-serial-test"))
      {
	int result = 0;            
	int port = 0;
	int command = 0;
	int txlen = 0;
	int rxlen = 0;
	char txbuf [10];
	char rxbuf [10];
      
	if (!client->serial (&result, port, command, txlen, txbuf, rxlen, rxbuf))
	  fail (argv [i - 1]);
      
	printf ("result value is %d, contents of buffers depends on you\n", result);
      }
      else if (!strcmp (argv [i], "-set-web-password"))
      {
	if (!client->setWebPassword (i + 2 < argc ? argv [i + 1] : 0, i + 2 < argc ? argv [i + 2] : 0))
	  fail (argv [i]);
	i+=2;
      }
      else if (!strcmp (argv [i], "-lock"))
      {
	client->disableTask ("tSnmpd");
	client->disableTask ("flash");
	client->disableTask ("ucflash");
	client->disableTask ("test");
      }
      else if (!strcmp (argv [i], "-disable"))
      {
	if (++i == argc)
	{
	  printf ("expected taskname\n");
	  usage (0);
	}
	if (!client->disableTask (argv [i]))
	  fail (argv [i- 1]);
      }
      else if (!strcmp (argv [i], "-get-packet-counts"))
      {
	char s1 [40],s2 [40];
	m3_int64*good,*bad;
	int numPorts;
	
	if (!(numPorts = client->getPacketCounters (&good, &bad)))
	  fail (argv [i- 1]);

	for (int j = 0; j < numPorts; j++)
	{
	  printf ("%d. %s/%s\n", j, string64 (s1, htonl (good [j].low), htonl (good [j].high)), string64 (s2, htonl (bad [j].low), htonl (bad [j].high)));
	}
      }
      else if (!strcmp (argv [i], "-notify"))
      {
	if (++i == argc)
	{
	  printf ("expected types field\n");
	  usage (0);
	}
	int types [2];
	if (sscanf (argv [i], "%x", &types [0]) != 1 || !types [0])
	{
	  printf ("just disabling notify.\n");
	  types [0] = types [1] = 0;
	}
	
	types [1] = 0;
	if (i + 1 < argc)
	  sscanf (argv [++i], "%x", &types [1]);
	
	char hostname [256];
	struct hostent*hp;
	m3_notify_message_t m;
	int length;
	struct sockaddr_in a;
	int fd;
	fd = socket (AF_INET, SOCK_DGRAM, 0);
	if (fd < 0)
	{
	  perror ("failed opening datagram socket");
	  fail (argv [i-1]);
	}
	memset ((char*)&a, 0, sizeof (a));

	a.sin_family = AF_INET;
	a.sin_addr.s_addr = INADDR_ANY;
	a.sin_port = 0;
      
	if (bind (fd, (struct sockaddr*) &a, sizeof (a)) < 0)
	{
	  perror ("binding datagram socket failed");
	  fail (argv [i-1]);
	}
      
	length = sizeof (a);
	if (getsockname (fd, (struct sockaddr*) &a, (socklen_t*) &length) < 0)
	{
	  perror ("getsockname failed");
	  fail (argv [i-1]);
	}
      
	if (gethostname (hostname, 255) < 0 || !(hp = gethostbyname (hostname)))
	{
	  printf ("couldn't get our ip address\n");
	  fail (argv [i-1]);
	}
      
	printf ("using udp port %d\n", ntohs (a.sin_port));
        
	if (numSwitches == 0 && !client->setNotify ((char*) hp->h_addr, ntohs (a.sin_port), types [0], types [1]))
	  fail (argv [i - 1]);

	for (int k = 0; k < numSwitches; k++)
	{

	  printf ("%d.%d.%d.%d\n",
		  (unsigned) (((char*)(&switches [k])) [0] & 0xff),
		  (unsigned) (((char*)(&switches [k])) [1] & 0xff),
		  (unsigned) (((char*)(&switches [k])) [2] & 0xff),
		  (unsigned) (((char*)(&switches [k])) [3] & 0xff));

	  if (!client->open (switches [k]) ||
	      !client->setNotify ((char*) hp->h_addr, ntohs (a.sin_port), types [0], types [1]))
	    fail (argv [i - 1]);
	}
	
	if (!types [0] && !types [1])
	  exit (0);
	
	for (;;)
	{
	  length = sizeof (a);      
	  while (recvfrom (fd, (char*) &m, sizeof (m), 0, (struct sockaddr*) &a, (socklen_t*) &length) == sizeof (m))
	  {
	    length = sizeof (a);
	    printf ("%d.%d.%d.%d ",
		    (unsigned) (((char*)(&a.sin_addr.s_addr)) [0] & 0xff),
		    (unsigned) (((char*)(&a.sin_addr.s_addr)) [1] & 0xff),
		    (unsigned) (((char*)(&a.sin_addr.s_addr)) [2] & 0xff),
		    (unsigned) (((char*)(&a.sin_addr.s_addr)) [3] & 0xff));
	
	    printf ("%s (%d) low: %d high: %d %s %d slot %d\n",
		    htonl (m.type) > 0 && htonl (m.type) <= M3_NUM_TRAPS ? 
		    m3_trap_names [htonl (m.type) - 1] : "?", 
		    htonl (m.type),
		    htonl (m.low), htonl (m.high), 
		    htonl (m.partType) > 0 && htonl (m.partType) <= M3_NUM_TABLES ? 
		    m3_table_names [htonl (m.partType) - 1] : "?",
		    htonl (m.index),
		    htonl (m.slot));
	  }
	}
      }    
      else if (!strcmp (argv [i], "-get-m2-packet-counts"))
      {
	int numPorts;
	
	if (!(numPorts = client->m2_getPacketCounters (m2_good, m2_bad)))
	  fail (argv [i]);
	
	printf ("xbar port good bad\n");
	for (int j = 0; j < numPorts; j++)
	{
	  printf ("%d %d %d %d\n", 1 + j / 16, 1 + j % 16, m2_good [j].low, m2_bad [j].low);
	}
      }
      else usage (argv [i]);
    }
  }
}
